/*
	CYD_ReceiveJPGWiFi.ino

	The purpose of this sketch is to receive a JPG image from c# application on the ESP32 and then display it to the CYD TFT.

*/

#include <FS.h>
#include <SPI.h>
#include <SD.h>
#include <ArduinoJson.h>
#include <WiFi.h>
#include <TFT_eSPI.h>
#include <JPEGDecoder.h>

/* data structures */
struct Config{
  char ssid[32];
  char password[29];
  char ep[32];
  char port[32];
};

//WiFi Configuration
Config config;
const char *cfg = "/etc/config.json";
const char *imageFile = "/images/tfer.jpg";
bool sdLoad = false;

WiFiServer wifiServer(80);
TFT_eSPI tft = TFT_eSPI();

void setup() {
  
  // Set all chip selects high to avoid bus contention during initialisation of each peripheral
  digitalWrite(15, HIGH);  // TFT screen chip select
  digitalWrite(5, HIGH);   // SD card chips select

  delay(3000);  //power-up safety delay
  
  Serial.begin(115200);
  tft.begin();
  
  tft.setRotation(1);  // landscape
  tft.fillScreen(TFT_BLACK);
  tft.setCursor(0, 0);
  tft.setTextFont(1);
  tft.setTextSize(2);
  tft.setTextColor(TFT_WHITE, TFT_BLACK);

  SD.begin();
  if(!SD.begin()){
    Serial.println("Card Mount Failure");
    sdLoad = false;
    return;
  }

  Serial.println("Initializing SD card ... ");
  if(!SD.begin()){
    Serial.println("ERROR - SD card initialization failed!");
    sdLoad = false;
    return;
  }
  
  File file = SD.open(cfg);
  if(!file){
    Serial.println("File doesn't exist");
  } else {
    Serial.println("Configuration file found");
    sdLoad = true;
  }
  file.close();

  if(sdLoad)
  {
      Serial.println(F("Loading configuration..."));
      loadConfiguration(config);
  }
  
  delay(1000);
  
  WiFi.begin(config.ssid, config.password);
  Serial.print("Connecting to WiFi");
  while (WiFi.status() != WL_CONNECTED) {
    delay(1000);
    Serial.print(".");
  }
  Serial.println();
  Serial.println("Connected to the WiFi network");
  Serial.println(WiFi.localIP());
  Serial.print("Port: ");
  Serial.println(config.port);
  
  tft.println("Connected to the WiFi network");
  tft.print("IP Address: ");
  tft.println(WiFi.localIP());
  tft.print("Port: ");
  tft.println(config.port);
 
  wifiServer.begin(atoi(config.port));
  Serial.println("Ready to connect.");
  tft.println("Ready to connect.");
  
  delay(3000); // wait 3 seconds
  displayImage();
}

void loop() {
  WiFiClient client = wifiServer.available();
  uint size=0;
  byte nString;
  File file;
  if (client) {
    while (client.connected()) {      
      file = SD.open(imageFile, FILE_WRITE);
      while ((size = client.available())>0) {
        //file.print(client.read());
		Serial.print("size = ");
		Serial.println(size);
		file.write(client.read());
      }      
      delay(10);       
    }
    file.close();
    Serial.println("File Closed");     
    client.stop();
	//now we need to send the image to display
	delay(1000);
	displayNewImage();
  }
}

void displayNewImage()
{
  tft.setRotation(2);  // portrait
  tft.fillScreen(TFT_BLACK);

  // The image is 300 x 300 pixels so we do some sums to position image in the middle of the screen!
  // Doing this by reading the image width and height from the jpeg info is left as an exercise!
  int x = (tft.width()  - 300) / 2 - 1;
  int y = (tft.height() - 300) / 2 - 1;

  drawSdJpeg(imageFile, x, y);     // This draws a jpeg pulled off the SD Card
}

void displayImage()
{
  tft.setRotation(2);  // portrait
  tft.fillScreen(TFT_BLACK);

  // The image is 300 x 300 pixels so we do some sums to position image in the middle of the screen!
  // Doing this by reading the image width and height from the jpeg info is left as an exercise!
  int x = (tft.width()  - 300) / 2 - 1;
  int y = (tft.height() - 300) / 2 - 1;

  drawSdJpeg("/images/cedric.jpg", x, y);     // This draws a jpeg pulled off the SD Card
}

void loadConfiguration(Config &config) {
  String data = "";
  File file = SD.open(cfg, FILE_READ);
  if (!file) {
    Serial.println("Failed to open file for reading");
    return;
  }
  while (file.available()) {
    data = file.readString();
  }
  file.close();

  JsonDocument doc;
  DeserializationError error = deserializeJson(doc, data);
  if (error) {
    Serial.print("deserializeJson() failed: ");
    Serial.println(error.c_str());
    return;
  }
  strlcpy(config.ssid, doc["SSID"], sizeof(config.ssid));
  strlcpy(config.password, doc["pass"], sizeof(config.password));
  strlcpy(config.ep, doc["ep"], sizeof(config.ep));
  strlcpy(config.port, doc["port"], sizeof(config.port));
}

void drawSdJpeg(const char *filename, int xpos, int ypos) {
  File jpegFile = SD.open( filename, FILE_READ);  // or, file handle reference for SD library
 
  if ( !jpegFile ) {
    Serial.print("ERROR: File \""); Serial.print(filename); Serial.println ("\" not found!");
    return;
  }

  Serial.println("===========================");
  Serial.print("Drawing file: "); Serial.println(filename);
  Serial.println("===========================");

  // Use one of the following methods to initialise the decoder:
  //bool decoded = JpegDec.decodeSdFile(jpegFile);  // Pass the SD file handle to the decoder,
  bool decoded = JpegDec.decodeSdFile(filename);  // or pass the filename (String or character array)

  if (decoded) {
    //jpegInfo();
    jpegRender(xpos, ypos);
  }
  else {
    Serial.println("Jpeg file format not supported!");
  }
}

void jpegRender(int xpos, int ypos) {
  uint16_t *pImg;
  uint16_t mcu_w = JpegDec.MCUWidth;
  uint16_t mcu_h = JpegDec.MCUHeight;
  uint32_t max_x = JpegDec.width;
  uint32_t max_y = JpegDec.height;

  bool swapBytes = tft.getSwapBytes();
  tft.setSwapBytes(true);
  
  // Jpeg images are draw as a set of image block (tiles) called Minimum Coding Units (MCUs)
  // Typically these MCUs are 16x16 pixel blocks
  // Determine the width and height of the right and bottom edge image blocks
  uint32_t min_w = jpg_min(mcu_w, max_x % mcu_w);
  uint32_t min_h = jpg_min(mcu_h, max_y % mcu_h);

  // save the current image block size
  uint32_t win_w = mcu_w;
  uint32_t win_h = mcu_h;

  // record the current time so we can measure how long it takes to draw an image
  uint32_t drawTime = millis();

  // save the coordinate of the right and bottom edges to assist image cropping
  // to the screen size
  max_x += xpos;
  max_y += ypos;

  // Fetch data from the file, decode and display
  while (JpegDec.read()) {    // While there is more data in the file
    pImg = JpegDec.pImage ;   // Decode a MCU (Minimum Coding Unit, typically a 8x8 or 16x16 pixel block)

    // Calculate coordinates of top left corner of current MCU
    int mcu_x = JpegDec.MCUx * mcu_w + xpos;
    int mcu_y = JpegDec.MCUy * mcu_h + ypos;

    // check if the image block size needs to be changed for the right edge
    if (mcu_x + mcu_w <= max_x) win_w = mcu_w;
    else win_w = min_w;

    // check if the image block size needs to be changed for the bottom edge
    if (mcu_y + mcu_h <= max_y) win_h = mcu_h;
    else win_h = min_h;

    // copy pixels into a contiguous block
    if (win_w != mcu_w)
    {
      uint16_t *cImg;
      int p = 0;
      cImg = pImg + win_w;
      for (int h = 1; h < win_h; h++)
      {
        p += mcu_w;
        for (int w = 0; w < win_w; w++)
        {
          *cImg = *(pImg + w + p);
          cImg++;
        }
      }
    }

    // calculate how many pixels must be drawn
    uint32_t mcu_pixels = win_w * win_h;

    // draw image MCU block only if it will fit on the screen
    if (( mcu_x + win_w ) <= tft.width() && ( mcu_y + win_h ) <= tft.height())
      tft.pushImage(mcu_x, mcu_y, win_w, win_h, pImg);
    else if ( (mcu_y + win_h) >= tft.height())
      JpegDec.abort(); // Image has run off bottom of screen so abort decoding
  }

  tft.setSwapBytes(swapBytes);
}